/* assemble.c   code generation for the Netwide Assembler
 *
 * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
 * Julian Hall. All rights reserved. The software is
 * redistributable under the licence given in the file "Licence"
 * distributed in the NASM archive.
 *
 * ARM instructions look like
 * 31 30 29 28 27 26 25 24 23 22 21 20
 *  C  O  N  D  -  -  -  -  -  -  -  -
 * 19 18 17 16 15 14 13 12 11 10  9  8
 *  -  -  -  -  -  -  -  -  -  -  -  -
 *  7  6  5  4  3  2  1  0
 *  -  -  -  -  -  -  -  -
 *
 * the actual codes (C syntax, i.e. octal):
 */

#include <stdio.h>
#include <string.h>

#include "nasm.h"
#include "nasmlib.h"
#include "assemble.h"
#include "insns.h"

#ifdef DEBUG
#include "insnsn.c"
extern int rt_debug;
#endif
extern int endianess;



extern struct itemplate *nasm_instructions[];

typedef struct
  {
    int sib_present;		       /* is a SIB byte necessary? */
    int bytes;			       /* # of bytes of offset needed */
    int size;			       /* lazy - this is sib+bytes+1 */
    unsigned char modrm, sib;	       /* the bytes themselves */
  } ea;

static efunc errfunc;
static struct ofmt *outfmt;
static ListGen *list;

static long calcsize (long, long, int, insn *, char *);
static void gencode (long, long, int, insn *, char *, long);
static int  shiftval (operand *o);
static int  regval (operand *o, int ordinary);
static int  matches (struct itemplate *, insn *);
static ea * process_ea (operand *, ea *, int, int, int);
static int  chsize (operand *, int);



/************************************************************************/
/* This routine wrappers the real output format's output routine,	*/
/* in order to pass a copy of the data off to the listing file		*/
/* generator at the same time.						*/
/************************************************************************/

static void out (long offset, long segto, void *data, unsigned long type,
		 long segment, long wrt) 
  {
    static long lineno;
    static char *lnfname;
    unsigned char p[4], *q = p;

    if ((type & OUT_TYPMASK) == OUT_ADDRESS)
      {
	if (segment != NO_SEG || wrt != NO_SEG)
	  {
	    /*
	     * This address is relocated. We must write it as
	     * OUT_ADDRESS, so there's no work to be done here.
	     */
	    list->output (offset, data, type);
	  }	 
	else
	  {
	    /*
	     * This is a non-relocated address, and we're going to
	     * convert it into RAWDATA format.
	     */
	    if ((type & OUT_SIZMASK) == 4)
	      {
		WRITELONG (q, * (long *) data);
		list->output (offset, p, OUT_RAWDATA+4);
	      } 
	    else
	      {
		WRITESHORT (q, * (long *) data);
		list->output (offset, p, OUT_RAWDATA+2);
	      }
	  }
      } 
    else if ((type & OUT_TYPMASK) == OUT_RAWDATA)
      {
	list->output (offset, data, type);
      } 
    else if ((type & OUT_TYPMASK) == OUT_RESERVE)
      {
	list->output (offset, NULL, type);
      } 
    else if ((type & OUT_TYPMASK) == OUT_REL2ADR ||
	     (type & OUT_TYPMASK) == OUT_REL3ADR ||
	     (type & OUT_TYPMASK) == OUT_REL4ADR)
      {
	list->output (offset, data, type);
      }

    
    if (src_get(&lineno,&lnfname))
      outfmt->current_dfmt->linenum(lnfname,lineno,segto);

    if (endianess != BIG_ENDIAN)
      {
	int size = (type & OUT_SIZMASK);
	if (size == 4)
	  {
	    q = &p[0];
	    WRITELONGB (q, *(long *) data);
	    outfmt->output (segto, p, type, segment, wrt);

	  }
	else
	  {
	    outfmt->output (segto, data, type, segment, wrt);
	  }
      }
    else
      {
	outfmt->output (segto, data, type, segment, wrt);
      }

  }

long assemble (long segment, long offset, int bits,
	       insn *instruction, struct ofmt *output, efunc error,
	       ListGen *listgen) 
{
    struct itemplate *temp;
    int    j;
    int    size_prob;
    long   insn_end;
    long   itimes;
    long   start = offset;
    long   wsize = 0;		       /* size for DB etc. */

    errfunc = error;		       /* to pass to other functions */
    outfmt = output;		       /* likewise */
    list = listgen;		       /* and again */

    switch (instruction->opcode) 
      {
	case   -1: return 0;
	case I_DCB: wsize = 1; break;
	case I_DCW: wsize = 2; break;
	case I_DCD: wsize = 4; break;
	case I_DQ: wsize = 8; break;
	case I_DT: wsize = 10; break;
      }

    if (wsize)
      {
	extop  * e;
	long   t = instruction->times;
	if (t < 0)
	    errfunc(ERR_PANIC, "instruction->times < 0 (%ld) in assemble()",t);

	while (t--) 		       /* repeat TIMES times */
	{
	    for (e = instruction->eops; e; e = e->next) 
	    {
		if (e->type == EOT_DB_NUMBER) 
		{
		    if (wsize == 1) {
			if (e->segment != NO_SEG)
			    errfunc (ERR_NONFATAL,
				     "one-byte relocation attempted");
			else {
			    unsigned char out_byte = e->offset;
			    out (offset, segment, &out_byte, OUT_RAWDATA+1,
				 NO_SEG, NO_SEG);
			}
		    } 
		    else if (wsize > 5) {
			errfunc (ERR_NONFATAL, "integer supplied to a D%c"
				 " instruction", wsize==8 ? 'Q' : 'T');
		    } 
		    else
			out (offset, segment, &e->offset,
			     OUT_ADDRESS+wsize, e->segment,
			     e->wrt);
		    offset += wsize;
		} 
		else if (e->type == EOT_DB_STRING) 
		{
		    int align;

		    out (offset, segment, e->stringval,
			 OUT_RAWDATA+e->stringlen, NO_SEG, NO_SEG);
		    align = e->stringlen % wsize;

		    if (align) {
			align = wsize - align;
			out (offset, segment, "\0\0\0\0\0\0\0\0",
			     OUT_RAWDATA+align, NO_SEG, NO_SEG);
			}
		    offset += e->stringlen + align;
		}
	    }
	    if (t > 0 && t == instruction->times-1) 
	    {
		/*
		 * Dummy call to list->output to give the offset to the
		 * listing module.
		 */
		list->output (offset, NULL, OUT_RAWDATA);
		list->uplevel (LIST_TIMES);
	    }
	}
	if (instruction->times > 1)
	    list->downlevel (LIST_TIMES);
	return offset - start;
    }

    if (instruction->opcode == I_INCBIN) 
    {
	static char fname[FILENAME_MAX];
	FILE        * fp;
	long        len;

	len = FILENAME_MAX-1;
	if (len > instruction->eops->stringlen)
	    len = instruction->eops->stringlen;
	strncpy (fname, instruction->eops->stringval, len);
	fname[len] = '\0';

	if ( (fp = fopen(fname, "rb")) == NULL)
	    error (ERR_NONFATAL, "`incbin': unable to open file `%s'", fname);
	else if (fseek(fp, 0L, SEEK_END) < 0)
	    error (ERR_NONFATAL, "`incbin': unable to seek on file `%s'",
		   fname);
	else 
	{
	    static char buf[2048];
	    long t = instruction->times;
	    long base = 0;

	    len = ftell (fp);
	    if (instruction->eops->next) {
		base = instruction->eops->next->offset;
		len -= base;
		if (instruction->eops->next->next &&
		    len > instruction->eops->next->next->offset)
		    len = instruction->eops->next->next->offset;
	    }
	    /*
	     * Dummy call to list->output to give the offset to the
	     * listing module.
	     */
	    list->output (offset, NULL, OUT_RAWDATA);
	    list->uplevel(LIST_INCBIN);
	    while (t--) 
	    {
		long l;

		fseek (fp, base, SEEK_SET);		
		l = len;
		while (l > 0) {
		    long m = fread (buf, 1, (l>sizeof(buf)?sizeof(buf):l),
				    fp);
		    if (!m) {
			/*
			 * This shouldn't happen unless the file
			 * actually changes while we are reading
			 * it.
			 */
			error (ERR_NONFATAL, "`incbin': unexpected EOF while"
			       " reading file `%s'", fname);
			t=0;  /* Try to exit cleanly */
			break;
		    }
		    out (offset, segment, buf, OUT_RAWDATA+m,
			 NO_SEG, NO_SEG);
		    l -= m;
		}
	    }
	    list->downlevel(LIST_INCBIN);
	    if (instruction->times > 1) {
		/*
		 * Dummy call to list->output to give the offset to the
		 * listing module.
		 */
		list->output (offset, NULL, OUT_RAWDATA);
		list->uplevel(LIST_TIMES);
		list->downlevel(LIST_TIMES);
	    }
	    fclose (fp);
	    return instruction->times * len;
	}
	return 0;		       /* if we're here, there's an error */
    }

    size_prob = FALSE;
    temp = nasm_instructions[instruction->opcode];

#ifdef DEBUG
    if (rt_debug)
      printf ("assemble:\n");
#endif
    
    while (temp->opcode != -1)
      {
	int m = matches (temp, instruction);

	if (m == 100) 		       /* matches! */
	  {
	    char *codes = temp->code;
	    long insn_size = 4; //calcsize(segment, offset, bits,
				//      instruction, codes);
	    itimes = instruction->times;
	    
	    if (insn_size < 0)	       /* shouldn't be, on pass two */
	      error (ERR_PANIC, "errors made it through from pass one");
	    else
	      while (itimes--)
		{
		  insn_end = offset + insn_size;
		  gencode (segment, offset, bits, instruction, codes, insn_end);
		  offset += insn_size;
		  if (itimes > 0 && itimes == instruction->times-1)
		    {
		      /*
		       * Dummy call to list->output to give the offset to the
		       * listing module.
		       */
		      list->output (offset, NULL, OUT_RAWDATA);
		      list->uplevel (LIST_TIMES);
		    }
		}
	      if (instruction->times > 1)
		list->downlevel (LIST_TIMES);
	      return (offset - start);
	  }
	temp++;
      }

    if (temp->opcode == -1) {	       /* didn't match any instruction */
	if (size_prob == 1)	       /* would have matched, but for size */
	    error (ERR_NONFATAL, "operation size not specified");
	else if (size_prob == 2)
	    error (ERR_NONFATAL, "mismatch in operand sizes");
	else
	    error (ERR_NONFATAL,
		   "invalid combination of opcode and operands");
    }
    return 0;
}


long insn_size (long segment, long offset, int bits,
		insn *instruction, efunc error) 
  {
    struct itemplate *temp;
#ifdef DEBUG
//    printf ("insn_size:\n");
#endif

    errfunc = error;		       /* to pass to other functions */

    if (instruction->opcode == -1)
    	return 0;

    if (instruction->opcode == I_DCB ||
	instruction->opcode == I_DCW ||
	instruction->opcode == I_DCD ||
	instruction->opcode == I_DQ ||
	instruction->opcode == I_DT) 
    {
	extop *e;
	long isize, osize, wsize = 0;  /* placate gcc */

	isize = 0;
	switch (instruction->opcode) 
	{
	  case I_DCB: wsize = 1; break;
	  case I_DCW: wsize = 2; break;
	  case I_DCD: wsize = 4; break;
	  case I_DQ: wsize = 8; break;
	  case I_DT: wsize = 10; break;
	}

	for (e = instruction->eops; e; e = e->next) 
	{
	    long align;

	    osize = 0;
	    if (e->type == EOT_DB_NUMBER)
		osize = 1;
	    else if (e->type == EOT_DB_STRING)
		osize = e->stringlen;

	    align = (-osize) % wsize;
	    if (align < 0)
		align += wsize;
	    isize += osize + align;
	}
	return isize * instruction->times;
    }

    if (instruction->opcode == I_INCBIN) 
      {
	char  fname[FILENAME_MAX];
	FILE  * fp;
	long  len;

	len = FILENAME_MAX-1;
	if (len > instruction->eops->stringlen)
	    len = instruction->eops->stringlen;
	strncpy (fname, instruction->eops->stringval, len);
	fname[len] = '\0';
	if ( (fp = fopen(fname, "rb")) == NULL )
	    error (ERR_NONFATAL, "`incbin': unable to open file `%s'", fname);
	else if (fseek(fp, 0L, SEEK_END) < 0)
	    error (ERR_NONFATAL, "`incbin': unable to seek on file `%s'",
		   fname);
	else 
	{
	    len = ftell (fp);
	    fclose (fp);
	    if (instruction->eops->next) 
	    {
		len -= instruction->eops->next->offset;
		if (instruction->eops->next->next &&
		    len > instruction->eops->next->next->offset)
		{
		    len = instruction->eops->next->next->offset;
		}
	    }
	    return instruction->times * len;
	}
	return 0;		       /* if we're here, there's an error */
      }

    temp = nasm_instructions[instruction->opcode];

    
    while (temp->opcode != -1)
      {
	if (matches(temp, instruction) == 100)
	  {
	    /* we've matched an instruction. */
	    long  isize;
	    char  * codes = temp->code;
	    int   j;

	    // All instructions are 4 bytes - RISC isn't it ?
	    
	    isize = 4; // calcsize(segment, offset, bits, instruction, codes);

	    return isize * instruction->times;
	  }
	temp++;
      }
    
    return -1;			       /* didn't match any instruction */
  }

static int imm_shift (int value)
  {
    unsigned int val = value;
    int shift = 16;
    int ret_val = -1;


    if (val <= 0xFF)
      {
	// To be compatible with MS's armasm (or is it Codemist's ?)
	// - makes it easier to test bit-patterns
	return (val);
      }
    
    // chop of the 2 LSBits as long as they are zeroes
    while ( (val & 0x03) == 0)
      {
	val >>= 2;
	shift--;
      }

    if (shift == 16)
      {
	shift = 0;
      }
    
    if ( val <= 0xFF)
      {
	ret_val = shift << 8;
	ret_val |= val;
      }

    return (ret_val);
  }
	

static void gencode (long segment, long offset, int bits,
		     insn *ins, char *codes, long insn_end) 
{
    int has_S_code;		/* S - setflag */
    int has_B_code;		/* B - setflag */
    int has_T_code;		/* T - setflag */
    int has_W_code;		/* ! => W flag */
    int has_F_code;		/* ^ => S flag */
    int keep;
    unsigned char c;
    unsigned char bytes[4];
    long          data, size;
    static int cc_code[] =	/* bit pattern of cc */
  {				/* order as enum in  */
    0x0E, 0x03, 0x02, 0x00,	/* nasm.h	     */
    0x0A, 0x0C, 0x08, 0x0D,
    0x09, 0x0B, 0x04, 0x01,
    0x05, 0x07, 0x06,
  };
#ifdef DEBUG
static char *CC[] =
  {				       /* condition code names */
    "AL", "CC", "CS", "EQ",
    "GE", "GT", "HI", "LE",
    "LS", "LT", "MI", "NE",
    "PL", "VC", "VS", "",
    "S"
};
#endif

    has_S_code = (ins->condition & C_SSETFLAG);
    has_B_code = (ins->condition & C_BSETFLAG);
    has_T_code = (ins->condition & C_TSETFLAG);
    has_W_code = (ins->condition & C_EXSETFLAG);
    has_F_code = (ins->condition & C_FSETFLAG);
    ins->condition = (ins->condition & 0x0F);
    
#ifdef DEBUG
    if (rt_debug)
      {
    printf ("gencode: instruction: %s%s", insn_names[ins->opcode],
	    CC[ins->condition & 0x0F]);
    if (has_S_code)
      printf ("S");
    if (has_B_code)
      printf ("B");
    if (has_T_code)
      printf ("T");
    if (has_W_code)
      printf ("!");
    if (has_F_code)
      printf ("^");

    printf ("\n");

    c = *codes;
    
    printf ("   (%d)  decode - '0x%02X'\n", ins->operands, c);


    bytes[0] = 0xB;
    bytes[1] = 0xE;
    bytes[2] = 0xE;
    bytes[3] = 0xF;
      }    
#endif
    // First condition code in upper nibble
    if (ins->condition < C_NONE)
      {
	c = cc_code[ins->condition] << 4;
      }
    else
      {
	c = cc_code[C_AL] << 4;	// is often ALWAYS but not always
      }


    switch (keep = *codes)
      {
	case 1:
	  // B, BL
	  ++codes;
	  c |= *codes++;
	  bytes[0] = c;

	  if (ins->oprs[0].segment != segment)
	    {
	      // fais une relocation
	      c = 1;
	      data = 0;	// Let the linker locate ?? 
	    }
	  else
	    {
	      c = 0;
	      data = ins->oprs[0].offset - (offset + 8);
	      
	      if (data % 4)
		{
		  errfunc (ERR_NONFATAL, "offset not aligned on 4 bytes");
		}
	    }
	  
	  if (data >= 0x1000)
	    {
	      errfunc (ERR_NONFATAL, "too long offset");
	    }

	  data = data >> 2;
	  bytes[1] = (data >> 16) & 0xFF;
	  bytes[2] = (data >> 8)  & 0xFF;
	  bytes[3] = (data )      & 0xFF;

	  if (c == 1)
	    {
//	      out (offset, segment, &bytes[0], OUT_RAWDATA+1, NO_SEG, NO_SEG);
	      out (offset, segment, &bytes[0], OUT_REL3ADR+4, ins->oprs[0].segment, NO_SEG);
	    }
	  else
	    {
	      out (offset, segment, &bytes[0], OUT_RAWDATA+4, NO_SEG, NO_SEG);
	    }
	  return;

	case 2:
	  // SWI
	  ++codes;
	  c |= *codes++;
	  bytes[0] = c;
	  data = ins->oprs[0].offset;
	  bytes[1] = (data >> 16) & 0xFF;
	  bytes[2] = (data >> 8) & 0xFF;
	  bytes[3] = (data) & 0xFF;
	  out (offset, segment, &bytes, OUT_RAWDATA+4, NO_SEG, NO_SEG);
	  return;
        case 3:
	  // BX
	  ++codes;
	  c |= *codes++;
	  bytes[0] = c;
	  bytes[1] = *codes++;
	  bytes[2] = *codes++;
	  bytes[3] = *codes++;
	  c = regval (&ins->oprs[0],1);
	  if (c == 15)	// PC
	    {
	      errfunc (ERR_WARNING, "'BX' with R15 has undefined behaviour");
	    }
	  else if (c > 15)
	    {
	      errfunc (ERR_NONFATAL, "Illegal register specified for 'BX'");
	    }

	  bytes[3] |= (c & 0x0F);
	  out (offset, segment, bytes, OUT_RAWDATA+4, NO_SEG, NO_SEG);
	  return;

        case 4:		// AND Rd,Rn,Rm
        case 5:		// AND Rd,Rn,Rm,<shift>Rs
        case 6:		// AND Rd,Rn,Rm,<shift>imm
        case 7:		// AND Rd,Rn,<shift>imm
	  ++codes;
#ifdef DEBUG
	  if (rt_debug)
	    {
	      printf ("         decode - '0x%02X'\n", keep);
	      printf ("           code - '0x%02X'\n", (unsigned char) (*codes));
	    }
#endif
	  bytes[0] = c | *codes;
	  ++codes;
	  
	  bytes[1] = *codes;
	  if (has_S_code)
	    bytes[1] |= 0x10;
	  c = regval (&ins->oprs[1],1);
	  // Rn in low nibble
	  bytes[1] |= c;

	  // Rd in high nibble
	  bytes[2] = regval (&ins->oprs[0],1) << 4;

	  if (keep != 7)
	    {
	      // Rm in low nibble
	      bytes[3] = regval (&ins->oprs[2],1);
	    }

	  // Shifts if any
	  if (keep == 5 || keep == 6)
	    {
	      // Shift in bytes 2 and 3
	      if (keep == 5)
		{
		  // Rs
		  c = regval (&ins->oprs[3],1);
		  bytes[2] |= c;

		  c = 0x10;		// Set bit 4 in byte[3]
		}
	      if (keep == 6)
		{
		  c = (ins->oprs[3].offset) & 0x1F;
		  
		  // #imm
		  bytes[2] |= c >> 1;
		  if (c & 0x01)
		    {
		      bytes[3] |= 0x80;
		    }
		  c = 0;		// Clr bit 4 in byte[3]
		}
	      // <shift>
	      c |= shiftval (&ins->oprs[3]) << 5;

	      bytes[3] |= c;
	    }
	  
	  // reg,reg,imm
	  if (keep == 7)
	    {
	      int shimm;
	      
	      shimm = imm_shift (ins->oprs[2].offset);

	      if (shimm == -1)
		{
		  errfunc (ERR_NONFATAL, "cannot create that constant");
		}
	      bytes[3] = shimm & 0xFF;
	      bytes[2] |= (shimm & 0xF00) >> 8;
	    }
	      
	  out (offset, segment, bytes, OUT_RAWDATA+4, NO_SEG, NO_SEG);
	  return;

        case 8:		// MOV Rd,Rm
        case 9:		// MOV Rd,Rm,<shift>Rs
        case 0xA:	// MOV Rd,Rm,<shift>imm
        case 0xB:	// MOV Rd,<shift>imm
	  ++codes;
#ifdef DEBUG
	  if (rt_debug)
	    {
	      printf ("         decode - '0x%02X'\n", keep);
	      printf ("           code - '0x%02X'\n", (unsigned char) (*codes));
	    }
#endif
	  bytes[0] = c | *codes;
	  ++codes;
	  
	  bytes[1] = *codes;
	  if (has_S_code)
	    bytes[1] |= 0x10;

	  // Rd in high nibble
	  bytes[2] = regval (&ins->oprs[0],1) << 4;

	  if (keep != 0x0B)
	    {
	      // Rm in low nibble
	      bytes[3] = regval (&ins->oprs[1],1);
	    }

	  // Shifts if any
	  if (keep == 0x09 || keep == 0x0A)
	    {
	      // Shift in bytes 2 and 3
	      if (keep == 0x09)
		{
		  // Rs
		  c = regval (&ins->oprs[2],1);
		  bytes[2] |= c;

		  c = 0x10;		// Set bit 4 in byte[3]
		}
	      if (keep == 0x0A)
		{
		  c = (ins->oprs[2].offset) & 0x1F;
		  
		  // #imm
		  bytes[2] |= c >> 1;
		  if (c & 0x01)
		    {
		      bytes[3] |= 0x80;
		    }
		  c = 0;		// Clr bit 4 in byte[3]
		}
	      // <shift>
	      c |= shiftval (&ins->oprs[2]) << 5;

	      bytes[3] |= c;
	    }
	  
	  // reg,imm
	  if (keep == 0x0B)
	    {
	      int shimm;
	      
	      shimm = imm_shift (ins->oprs[1].offset);

	      if (shimm == -1)
		{
		  errfunc (ERR_NONFATAL, "cannot create that constant");
		}
	      bytes[3] = shimm & 0xFF;
	      bytes[2] |= (shimm & 0xF00) >> 8;
	    }
	      
	  out (offset, segment, bytes, OUT_RAWDATA+4, NO_SEG, NO_SEG);
	  return;
	  

        case 0xC:	// CMP Rn,Rm
        case 0xD:	// CMP Rn,Rm,<shift>Rs
        case 0xE:	// CMP Rn,Rm,<shift>imm
        case 0xF:	// CMP Rn,<shift>imm
	  ++codes;

	  bytes[0] = c | *codes++;
	  
	  bytes[1] = *codes;

	  // Implicit S code
	  bytes[1] |= 0x10;

	  c = regval (&ins->oprs[0],1);
	  // Rn in low nibble
	  bytes[1] |= c;

	  // No destination
	  bytes[2] = 0;

	  if (keep != 0x0B)
	    {
	      // Rm in low nibble
	      bytes[3] = regval (&ins->oprs[1],1);
	    }

	  // Shifts if any
	  if (keep == 0x0D || keep == 0x0E)
	    {
	      // Shift in bytes 2 and 3
	      if (keep == 0x0D)
		{
		  // Rs
		  c = regval (&ins->oprs[2],1);
		  bytes[2] |= c;

		  c = 0x10;		// Set bit 4 in byte[3]
		}
	      if (keep == 0x0E)
		{
		  c = (ins->oprs[2].offset) & 0x1F;
		  
		  // #imm
		  bytes[2] |= c >> 1;
		  if (c & 0x01)
		    {
		      bytes[3] |= 0x80;
		    }
		  c = 0;		// Clr bit 4 in byte[3]
		}
	      // <shift>
	      c |= shiftval (&ins->oprs[2]) << 5;

	      bytes[3] |= c;
	    }
	  
	  // reg,imm
	  if (keep == 0x0F)
	    {
	      int shimm;
	      
	      shimm = imm_shift (ins->oprs[1].offset);

	      if (shimm == -1)
		{
		  errfunc (ERR_NONFATAL, "cannot create that constant");
		}
	      bytes[3] = shimm & 0xFF;
	      bytes[2] |= (shimm & 0xF00) >> 8;
	    }
	      
	  out (offset, segment, bytes, OUT_RAWDATA+4, NO_SEG, NO_SEG);
	  return;
	  
        case 0x10:	// MRS Rd,<psr>
	  ++codes;

	  bytes[0] = c | *codes++;
	  
	  bytes[1] = *codes++;

	  // Rd
	  c = regval (&ins->oprs[0],1);

	  bytes[2] = c << 4;

	  bytes[3] = 0;

	  c = ins->oprs[1].basereg;

	  if (c == R_CPSR || c == R_SPSR)
	    {
	      if (c == R_SPSR)
		{
		  bytes[1] |= 0x40;
		}
	    }
	  else
	    {
	      errfunc (ERR_NONFATAL, "CPSR or SPSR expected");
	    }

	  out (offset, segment, bytes, OUT_RAWDATA+4, NO_SEG, NO_SEG);

	  return;
	  
        case 0x11:	// MSR <psr>,Rm
	case 0x12:	// MSR <psrf>,Rm
        case 0x13:	// MSR <psrf>,#expression
	  ++codes;

	  bytes[0] = c | *codes++;
	  
	  bytes[1] = *codes++;

	  bytes[2] = *codes;

 
	  if (keep == 0x11 || keep == 0x12)
	    {
	      // Rm
	      c = regval (&ins->oprs[1],1);

	      bytes[3] = c;
	    }
	  else
	    {
	      int shimm;
	      
	      shimm = imm_shift (ins->oprs[1].offset);

	      if (shimm == -1)
		{
		  errfunc (ERR_NONFATAL, "cannot create that constant");
		}
	      bytes[3] = shimm & 0xFF;
	      bytes[2] |= (shimm & 0xF00) >> 8;
	    }
	      
	  c = ins->oprs[0].basereg;

	  if ( keep == 0x11)
	    {
	      if ( c == R_CPSR || c == R_SPSR)
		{
		if ( c== R_SPSR)
		  {
		    bytes[1] |= 0x40;
		  }
		}
	    else
	      {
		errfunc (ERR_NONFATAL, "CPSR or SPSR expected");
	      }
	    }
	  else
	    {
	      if ( c == R_CPSR_FLG || c == R_SPSR_FLG)
		{
		  if ( c== R_SPSR_FLG)
		    {
		      bytes[1] |= 0x40;
		    }
		}
	      else
		{
		  errfunc (ERR_NONFATAL, "CPSR_flg or SPSR_flg expected");
		}
	    }
	  break;

        case 0x14:	// MUL  Rd,Rm,Rs
        case 0x15:	// MULA Rd,Rm,Rs,Rn
	  ++codes;
	  
	  bytes[0] = c | *codes++;
	  
	  bytes[1] = *codes++;

	  bytes[3] = *codes;

	  // Rd
	  bytes[1] |= regval (&ins->oprs[0],1);
	  if (has_S_code)
	    bytes[1] |= 0x10;

	  // Rm
	  bytes[3] |= regval (&ins->oprs[1],1);

	  // Rs
	  bytes[2] = regval (&ins->oprs[2],1);

	  if (keep == 0x15)
	    {
	      bytes[2] |= regval (&ins->oprs[3],1) << 4;
	    }
	  break;

        case 0x16:	// SMLAL RdHi,RdLo,Rm,Rs
	  ++codes;
	  
	  bytes[0] = c | *codes++;
	  
	  bytes[1] = *codes++;

	  bytes[3] = *codes;

	  // RdHi
	  bytes[1] |= regval (&ins->oprs[1],1);
	  if (has_S_code)
	    bytes[1] |= 0x10;

	  // RdLo
	  bytes[2] = regval (&ins->oprs[0],1) << 4;
	  // Rm
	  bytes[3] |= regval (&ins->oprs[2],1);

	  // Rs
	  bytes[2] |= regval (&ins->oprs[3],1);

	  break;
	  
        case 0x17:	// LDR Rd, expression
	  ++codes;
	  
	  bytes[0] = c | *codes++;
	  
	  bytes[1] = *codes++;

	  // Rd
	  bytes[2] = regval (&ins->oprs[0],1) << 4;
	  if (has_B_code)
	    bytes[1] |= 0x40;
	  if (has_T_code)
	    {
	      errfunc (ERR_NONFATAL, "'T' not allowed in pre-index mode");
	    }
	  if (has_W_code)
	    {
	      errfunc (ERR_NONFATAL, "'!' not allowed");
	    }

	  // Rn - implicit R15
	  bytes[1] |= 0xF;

	  if (ins->oprs[1].segment != segment)
	    {
	      errfunc (ERR_NONFATAL, "label not in same segment");
	    }
	  
	  data = ins->oprs[1].offset - (offset + 8);

	  if (data < 0)
	    {
	      data = -data;
	    }
	  else
	    {
	      bytes[1] |= 0x80;
	    }

	  if (data >= 0x1000)
	    {
	      errfunc (ERR_NONFATAL, "too long offset");
	    }

	  bytes[2] |= ((data & 0xF00) >> 8);
	  bytes[3] = data & 0xFF;
	  break;
	  
        case 0x18:	// LDR Rd, [Rn]
	  ++codes;
	  
	  bytes[0] = c | *codes++;
	  
	  bytes[1] = *codes++;

	  // Rd
	  bytes[2] = regval (&ins->oprs[0],1) << 4;
	  if (has_B_code)
	    bytes[1] |= 0x40;
	  if (has_T_code)
	    {
	      bytes[1] |= 0x20;		// write-back
	    }
	  else
	    {
	      bytes[0] |= 0x01;		// implicit pre-index mode
	    }

	  if (has_W_code)
	    {
	      bytes[1] |= 0x20;		// write-back
	    }

	  // Rn
	  c = regval (&ins->oprs[1],1);
	  bytes[1] |= c;

	  if (c == 0x15)		// R15
	    data = -8;
	  else
	    data = 0;

	  if (data < 0)
	    {
	      data = -data;
	    }
	  else
	    {
	      bytes[1] |= 0x80;
	    }

	  bytes[2] |= ((data & 0xF00) >> 8);
	  bytes[3] = data & 0xFF;
	  break;
	  
        case 0x19:	// LDR Rd, [Rn,#expression]
	case 0x20:	// LDR Rd, [Rn,Rm]
	case 0x21:	// LDR Rd, [Rn,Rm,shift]
	  ++codes;
	  
	  bytes[0] = c | *codes++;
	  
	  bytes[1] = *codes++;

	  // Rd
	  bytes[2] = regval (&ins->oprs[0],1) << 4;
	  if (has_B_code)
	    bytes[1] |= 0x40;

	  // Rn
	  c = regval (&ins->oprs[1],1);
	  bytes[1] |= c;

	  if (ins->oprs[ins->operands-1].bracket)	// FIXME: Bracket on last operand -> pre-index  <--
	    {
	      bytes[0] |= 0x01;		// pre-index mode
	      if (has_W_code)
		{
		  bytes[1] |= 0x20;
		}
	      if (has_T_code)
		{
		  errfunc (ERR_NONFATAL, "'T' not allowed in pre-index mode");
		}
	    }
	  else
	    {
	      if (has_T_code)		// Forced write-back in post-index mode
		{
		  bytes[1] |= 0x20;
		}
	      if (has_W_code)
		{
		  errfunc (ERR_NONFATAL, "'!' not allowed in post-index mode");
		}
	    }
	  
	  if (keep == 0x19)
	    {
	      data = ins->oprs[2].offset;

	      if (data < 0)
		{
		  data = -data;
		}
	      else
		{
		  bytes[1] |= 0x80;
		}

	      if (data >= 0x1000)
		{
		  errfunc (ERR_NONFATAL, "too long offset");
		}
	  
	      bytes[2] |= ((data & 0xF00) >> 8);
	      bytes[3] = data & 0xFF;
	    }
	  else 
	    {
	      if (ins->oprs[2].minus == 0)
		{
		  bytes[1] |= 0x80;
		}
	      c = regval (&ins->oprs[2],1);
	      bytes[3] = c;

	      if (keep == 0x21)
		{
		  c = ins->oprs[3].offset;
		  if (c > 0x1F)
		    {
		      errfunc (ERR_NONFATAL, "too large shiftvalue");
		      c = c & 0x1F;
		    }
		      
		  bytes[2] |= c >> 1;
		  if (c & 0x01)
		    {
		      bytes[3] |= 0x80;
		    }
		  bytes[3] |= shiftval (&ins->oprs[3]) << 5;
		}
	    }
	      
	  break;
	  
        case 0x22:	// LDRH Rd, expression
	  ++codes;
	  
	  bytes[0] = c | 0x01;		// Implicit pre-index
	  
	  bytes[1] = *codes++;

	  // Rd
	  bytes[2] = regval (&ins->oprs[0],1) << 4;
	  
	  // Rn - implicit R15
	  bytes[1] |= 0xF;

	  if (ins->oprs[1].segment != segment)
	    {
	      errfunc (ERR_NONFATAL, "label not in same segment");
	    }
	  
	  data = ins->oprs[1].offset - (offset + 8);

	  if (data < 0)
	    {
	      data = -data;
	    }
	  else
	    {
	      bytes[1] |= 0x80;
	    }

	  if (data >= 0x100)
	    {
	      errfunc (ERR_NONFATAL, "too long offset");
	    }
	  bytes[3] = *codes++;

	  bytes[2] |= ((data & 0xF0) >> 4);
	  bytes[3] |= data & 0xF;
	  break;
	  
        case 0x23:	// LDRH Rd, Rn
	  ++codes;
	  
	  bytes[0] = c | 0x01;		// Implicit pre-index
	  
	  bytes[1] = *codes++;

	  // Rd
	  bytes[2] = regval (&ins->oprs[0],1) << 4;
	  
	  // Rn
	  c = regval (&ins->oprs[1],1);
	  bytes[1] |= c;

	  if (c == 0x15)		// R15
	    data = -8;
	  else
	    data = 0;
	  
	  if (data < 0)
	    {
	      data = -data;
	    }
	  else
	    {
	      bytes[1] |= 0x80;
	    }

	  if (data >= 0x100)
	    {
	      errfunc (ERR_NONFATAL, "too long offset");
	    }
	  bytes[3] = *codes++;

	  bytes[2] |= ((data & 0xF0) >> 4);
	  bytes[3] |= data & 0xF;
	  break;
	  
        case 0x24:	// LDRH Rd, Rn, expression
        case 0x25:	// LDRH Rd, Rn, Rm
	  ++codes;
	  
	  bytes[0] = c;
	  
	  bytes[1] = *codes++;

	  // Rd
	  bytes[2] = regval (&ins->oprs[0],1) << 4;

	  // Rn
	  c = regval (&ins->oprs[1],1);
	  bytes[1] |= c;

	  if (ins->oprs[ins->operands-1].bracket)	// FIXME: Bracket on last operand -> pre-index  <--
	    {
	      bytes[0] |= 0x01;		// pre-index mode
	      if (has_W_code)
		{
		  bytes[1] |= 0x20;
		}
	    }
	  else
	    {
	      if (has_W_code)
		{
		  errfunc (ERR_NONFATAL, "'!' not allowed in post-index mode");
		}
	    }

	  bytes[3] = *codes++;

	  if (keep == 0x24)
	    {
	      data = ins->oprs[2].offset;

	      if (data < 0)
		{
		  data = -data;
		}
	      else
		{
		  bytes[1] |= 0x80;
		}
	      
	      if (data >= 0x100)
		{
		  errfunc (ERR_NONFATAL, "too long offset");
		}

	      bytes[2] |= ((data & 0xF0) >> 4);
	      bytes[3] |= data & 0xF;
	    }
	  else 
	    {
	      if (ins->oprs[2].minus == 0)
		{
		  bytes[1] |= 0x80;
		}
	      c = regval (&ins->oprs[2],1);
	      bytes[3] |= c;

	    }
	  break;
	  
        case 0x26:	// LDM/STM Rn, {reg-list}
	  ++codes;
	  
	  bytes[0] = c;

	  bytes[0] |= (*codes >> 4) & 0xF;
	  bytes[1] = (*codes << 4) & 0xF0;
	  ++codes;

	  if (has_W_code)
	    {
	      bytes[1] |= 0x20;
	    }
	  if (has_F_code)
	    {
	      bytes[1] |= 0x40;
	    }
	    
	  // Rn
	  bytes[1] |= regval (&ins->oprs[0],1);

	  data = ins->oprs[1].basereg;

	  bytes[2] = ((data >> 8) & 0xFF);
	  bytes[3] = (data & 0xFF);
	  
	  break;
	  
        case 0x27:	// SWP Rd, Rm, [Rn]
	  ++codes;
	  
	  bytes[0] = c;

	  bytes[0] |= *codes++;
	  
	  bytes[1] = regval (&ins->oprs[2],1);
	  if (has_B_code)
	    {
	      bytes[1] |= 0x40;
	    }
	  bytes[2] = regval (&ins->oprs[0],1) << 4;
	  bytes[3] = *codes++;
	  bytes[3] |= regval (&ins->oprs[1],1);
	  break;
	  
        default:
	  errfunc (ERR_FATAL, "unknown decoding of instruction");

	  bytes[0] = c;
	  // And a fix nibble
	  ++codes;
	  bytes[0] |= *codes++;

	 if (*codes == 0x01)		// An I bit
	   {

	   }
	 if (*codes == 0x02)		// An I bit
	   {

	   }
	 ++codes;
      }
    out (offset, segment, bytes, OUT_RAWDATA+4, NO_SEG, NO_SEG);
}


static int shiftval (operand *o)
  {
    switch (o->shiftflag)
      {
	case S_ASR:
	  return (2);
	case S_LSL:
	  return (0);
	  
	case S_LSR:
	  return (1);
	case S_ROR:
	  return (3);
	case S_RRX:
	default:
	  return (23456);
	}
  }

      
static int regval (operand *o, int ordinary) 
{
    switch (o->basereg)
      {
	case R_0:
	  return 0;
	case R_1:
	  return 1;
	case R_2:
	  return 2;
	case R_3:
	  return 3;
	case R_4:
	  return 4;
	case R_5:
	  return 5;
	case R_6:
	  return 6;
	case R_7:
	  return 7;
	case R_8:
	  return 8;
	case R_9:
	  return 9;
	case R_10:
	case R_SL:
	  return 10;
	case R_11:
	case R_FP:
	  return 11;
	case R_12:
	case R_IP:
	  return 12;
	case R_13:
	case R_SP:
	  return 13;
	case R_14:
	case R_LR:
	  return 14;
	case R_15:
	case R_PC:
	  return 15;
	case R_CPSR:
	case R_CPSR_FLG:
	  if (ordinary)
	    {
	      errfunc (ERR_NONFATAL, "CSPR not a valid register");
	    }
	  return 16;
	case R_SPSR:
	case R_SPSR_FLG:
	  if (ordinary)
	    {
	      errfunc (ERR_PANIC, "SPSR not a valid register");
	    }
	  return 17;
	default:			       /* panic */
	  errfunc (ERR_PANIC, "invalid register operand given to regval()");
	  return 0;
    }
}


static int matches (struct itemplate *itemp, insn *instruction) 
{
    int i, size[3], asize, oprs, ret;

    ret = 100;

    /*
     * Check the opcode
     */
    if (itemp->opcode != instruction->opcode) return 0;

    /*
     * Count the operands
     */
    if (itemp->operands != instruction->operands)
      {
#ifdef DEBUG
    if (rt_debug)
      {
	printf ("Instruction: %s\n", insn_names[instruction->opcode]);
	printf ("Wrong no of operands expected %d (got %d)\n", itemp->operands, instruction->operands);
      }
#endif
	return 0;
      }
	
#if 0
    /*
     * Check that no spurious colons or TOs are present
     */
    for (i=0; i<itemp->operands; i++)
	if (instruction->oprs[i].type & ~itemp->opd[i] & (COLON|TO))
	    return 0;
#endif
    /*
     * Check that the operand flags all match up
     */

    for (i=0; i<itemp->operands; i++)
      {

	if (itemp->opd[i] & ~instruction->oprs[i].type)
	  {
#ifdef DEBUG
	    if (rt_debug)
	      {
		printf ("Instruction: %s\n", insn_names[instruction->opcode]);
		printf ("Operands don't match, want 0x%04X have 0x%04X\n",
			itemp->opd[i], instruction->oprs[i].type);
	      }
#endif
	    return 0;
	  }
      }
    
    /*
     * Check operand sizes
     */
      oprs = itemp->operands;
      
    return ret;
}

static ea *process_ea (operand *input, ea *output, int addrbits, int rfield,
		       int forw_ref) 
{
    if (!(REGISTER & ~input->type)) {  /* it's a single register */
	static int regs[] = {
	  R_0,   R_1,   R_2,   R_3,   R_4,   R_5,   R_6,   R_7,
	  R_8,   R_9,   R_10,   R_11,   R_12,   R_13,   R_14,   R_15,
	  R_CPSR,  R_SPSR,
	};
	int i;

	for (i=0; i<elements(regs); i++)
	    if (input->basereg == regs[i]) break;
	if (i<elements(regs)) {
	    output->sib_present = FALSE;/* no SIB necessary */
	    output->bytes = 0;	       /* no offset necessary either */
            output->modrm = 0xC0 | (rfield << 3) | (i & 7);
	} 
	else
	    return NULL;
    } else {			       /* it's a memory reference */
	if (input->basereg==-1 && (input->indexreg==-1 || input->scale==0)) {
	    /* it's a pure offset */
	    if (input->addr_size)
		addrbits = input->addr_size;
	    output->sib_present = FALSE;
	    output->bytes = (addrbits==32 ? 4 : 2);
	    output->modrm = (addrbits==32 ? 5 : 6) | (rfield << 3);
	} 
	else {		       /* it's an indirection */
	    int i=input->indexreg, b=input->basereg, s=input->scale;
	    long o=input->offset, seg=input->segment;
	    int hb=input->hintbase, ht=input->hinttype;
	    int t;

	    if (s==0) i = -1;	       /* make this easy, at least */

	    if (i==R_0 || i==R_1 || i==R_2 || i==R_3
		|| i==R_4 || i==R_5 || i==R_6 || i==R_7
		|| b==R_8 || b==R_9 || b==R_10 || b==R_11
		|| b==R_12 || b==R_13 || b==R_14 || b==R_15) {
		/* it must be a 32-bit memory reference. Firstly we have
		 * to check that all registers involved are type Exx. */
		if (i!=-1 && i!=R_0)
		    return NULL;
		if (b!=-1 && b!=R_1)
		    return NULL;

		/* While we're here, ensure the user didn't specify WORD. */
		if (input->addr_size == 16)
		    return NULL;


		if (i==-1 ) {/* no SIB needed */
		    int mod, rm;
		    switch(b) {
		      case R_0: rm = 0; break;
		      case R_1: rm = 1; break;
		      case R_2: rm = 2; break;
		      case R_3: rm = 3; break;
		      case R_4: rm = 5; break;
		      case R_5: rm = 6; break;
		      case R_6: rm = 7; break;
		      case -1: rm = 5; break;
		      default:	       /* should never happen */
			return NULL;
		    }
		    if (b==-1 )
		    	mod = 0;
		    else if (input->eaflags & EAF_BYTEOFFS ||
			     (o>=-128 && o<=127 && seg==NO_SEG && !forw_ref &&
			      !(input->eaflags & EAF_WORDOFFS))) {
		    	mod = 1;
		    } 
		    else
		    	mod = 2;

		    output->sib_present = FALSE;
		    output->bytes = (b==-1 || mod==2 ? 4 : mod);
		    output->modrm = (mod<<6) | (rfield<<3) | rm;
		} 
		else {	       /* we need a SIB */
		    int mod, scale, index, base;

		    switch (b) {
		      case R_0: base = 0; break;
		      case R_1: base = 1; break;
		      case R_2: base = 2; break;
		      case R_3: base = 3; break;
		      case R_4: base = 4; break;
		      case R_5: case -1: base = 5; break;
		      case R_6: base = 6; break;
		      case R_7: base = 7; break;
		      default:	       /* then what the smeg is it? */
			return NULL;  /* panic */
		    }

		    switch (i) {
		      case R_0: index = 0; break;
		      case R_1: index = 1; break;
		      case R_2: index = 2; break;
		      case R_3: index = 3; break;
		      case -1: index = 4; break;
		      case R_5: index = 5; break;
		      case R_6: index = 6; break;
		      case R_7: index = 7; break;
		      default:	       /* then what the smeg is it? */
			return NULL;  /* panic */
		    }

		    if (i==-1) s = 1;
		    switch (s) {
		      case 1: scale = 0; break;
		      case 2: scale = 1; break;
		      case 4: scale = 2; break;
		      case 8: scale = 3; break;
		      default:	       /* then what the smeg is it? */
			return NULL;  /* panic */
		    }

		    if (b==-1 || (b!=R_LR && o==0 &&
				  seg==NO_SEG && !forw_ref &&
				  !(input->eaflags &
				    (EAF_BYTEOFFS|EAF_WORDOFFS))))
		    	mod = 0;
		    else if (input->eaflags & EAF_BYTEOFFS ||
			     (o>=-128 && o<=127 && seg==NO_SEG && !forw_ref &&
			      !(input->eaflags & EAF_WORDOFFS)))
		    	mod = 1;
		    else
		    	mod = 2;

		    output->sib_present = TRUE;
		    output->bytes = (b==-1 || mod==2 ? 4 : mod);
		    output->modrm = (mod<<6) | (rfield<<3) | 4;
		    output->sib = (scale<<6) | (index<<3) | base;
		}
	    } 
	    else {		       /* it's 16-bit */
		int mod, rm;

		/* check all registers are BX, BP, SI or DI */
		if ((b!=-1 && b!=R_SP && b!=R_0) ||
		    (i!=-1 && i!=R_SP && i!=R_0))
		    return NULL;

		/* ensure the user didn't specify DWORD */
		if (input->addr_size == 32)
		    return NULL;

		if (s!=1 && i!=-1) return NULL;/* no can do, in 16-bit EA */
		if (b==-1 && i!=-1) b ^= i ^= b ^= i;   /* swap them round */
		if (b==i) return NULL;/* shouldn't ever happen, in theory */
		if (b==-1)	       /* pure offset: handled above */
		    return NULL;      /* so if it gets to here, panic! */

		rm = -1;
		if (rm==-1)	       /* can't happen, in theory */
		    return NULL;      /* so panic if it does */

		if (o==0 && seg==NO_SEG && !forw_ref && rm!=6 &&
		    !(input->eaflags & (EAF_BYTEOFFS|EAF_WORDOFFS)))
		    mod = 0;
		else if (input->eaflags & EAF_BYTEOFFS ||
			 (o>=-128 && o<=127 && seg==NO_SEG && !forw_ref &&
			  !(input->eaflags & EAF_WORDOFFS)))
		    mod = 1;
		else
		    mod = 2;

		output->sib_present = FALSE;  /* no SIB - it's 16-bit */
		output->bytes = mod;  /* bytes of offset needed */
		output->modrm = (mod<<6) | (rfield<<3) | rm;
	    }
	}
    }
    output->size = 1 + output->sib_present + output->bytes;
    return output;
}

static int chsize (operand *input, int addrbits) 
{
    if (!(MEMORY & ~input->type)) {
	int i=input->indexreg, b=input->basereg;

	if (input->scale==0) i = -1;

	if (i == -1 && b == -1) /* pure offset */
	    return (input->addr_size != 0 && input->addr_size != addrbits);

	return (addrbits==32);
    } 
    else
    	return 0;
}
