/* parser.c   source line parser for the Netwide Assembler
 *
 * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
 * Julian Hall. All rights reserved. The software is
 * redistributable under the licence given in the file "Licence"
 * distributed in the NASM archive.
 *
 * initial version 27/iii/95 by Simon Tatham
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <ctype.h>

#include "nasm.h"
#include "nasmlib.h"
#include "parser.h"
#include "float.h"



static int is_comma_next (void);

static int i;
static struct tokenval tokval;
static efunc error;
static struct ofmt *outfmt;  /* Structure of addresses of output routines */
static loc_t *location;	     /* Pointer to current line's segment,offset */

static int regtoval (int reg)
  {
    int val = 0;
    switch (reg)
      {
	case R_0:
	  val = 1;
	  break;
	case R_1:
	  val = 1 << 1;
	  break;
	case R_2:
	  val = 1 << 2;
	  break;
	case R_3:
	  val = 1 << 3;
	  break;
	case R_4:
	  val = 1 << 4;
	  break;
	case R_5:
	  val = 1 << 5;
	  break;
	case R_6:
	  val = 1 << 6;
	  break;
	case R_7:
	  val = 1 << 7;
	  break;
	case R_8:
	  val = 1 << 8;
	  break;
	case R_9:
	  val = 1 << 9;
	  break;
	case R_10:
	case R_SL:	  
	  val = 1 << 10;
	  break;
	case R_11:
	case R_FP:	  
	  val = 1 << 11;
	  break;
	case R_12:
	case R_IP:	  
	  val = 1 << 12;
	  break;
	case R_13:
	case R_SP:	  
	  val = 1 << 13;
	  break;
	case R_14:
	case R_LR:	  
	  val = 1 << 14;
	  break;
	case R_15:
	case R_PC:	  
	  val = 1 << 15;
	  break;
	default:
	error (ERR_NONFATAL, "illegal register in register-list");
	  val = 0;
      }

    return (val);
  }


void parser_global_info (struct ofmt *output, loc_t *locp) 
{
    outfmt = output;
    location = locp;
}

insn *parse_line (int pass, char *buffer, insn *result,
		  efunc errfunc, evalfunc evaluate, ldfunc ldef) 
{
    int operand;
    int critical;
    struct eval_hints hints;
    int bracket;
    int minus;

    result->forw_ref = FALSE;
    error = errfunc;

    stdscan_reset();
    stdscan_bufptr = buffer;
    i = stdscan(NULL, &tokval);

    result->label = NULL;	       /* Assume no label */
    result->eops = NULL;	       /* must do this, whatever happens */
    result->operands = 0;	       /* must initialise this */

    if (i==0)
      {				       /* blank line - ignore */
	result->opcode = -1;	       /* and no instruction either */
	return result;
      }
    if (i != TOKEN_ID && i != TOKEN_INSN && i != TOKEN_PREFIX &&
	(i!=TOKEN_REG))
      {
	error (ERR_NONFATAL, "label or instruction expected"
	       " at start of line");
	result->opcode = -1;
	return result;
      }

    if (i == TOKEN_ID)
      {	       /* there's a label here */
	result->label = tokval.t_charptr;
	i = stdscan(NULL, &tokval);
	if (i == ':')
	  {		       /* skip over the optional colon */
	    i = stdscan(NULL, &tokval);
	  }
	else if (i == 0)
	  {
	    error (ERR_WARNING|ERR_WARN_OL|ERR_PASS1,
		   "label alone on a line without a colon might be in error");
	  }
	if (i != TOKEN_INSN || tokval.t_integer != I_EQU)
	  {
	    /*
	     * FIXME: location->segment could be NO_SEG, in which case
	     * it is possible we should be passing 'abs_seg'. Look into this.
	     * Work out whether that is *really* what we should be doing.
	     * Generally fix things. I think this is right as it is, but
	     * am still not certain.
	     */
	    ldef (result->label, location->segment,
		  location->offset, NULL, TRUE, FALSE, outfmt, errfunc);
	  }
      }

    if (i==0)
      {
	result->opcode = -1;	       /* this line contains just a label */
	return result;
      }

    result->nprefix = 0;
    result->times = 1L;

    while (i == TOKEN_PREFIX || i==TOKEN_REG)
      {
	/*
	 * Handle special case: the TIMES prefix.
	 */
	if (i == TOKEN_PREFIX && tokval.t_integer == P_TIMES)
	  {
	    expr *value;

	    i = stdscan(NULL, &tokval);
	    value = evaluate (stdscan, NULL, &tokval, NULL, pass, error, NULL);
	    i = tokval.t_type;
	    if (!value)
	      {			       /* but, error in evaluator */
		result->opcode = -1;   /* unrecoverable parse error: */
		return result;	       /* ignore this instruction */
	      }
	    if (!is_simple (value))
	      {
		error (ERR_NONFATAL,
		       "non-constant argument supplied to TIMES");
		result->times = 1L;
	      }
	    else
	      {
		result->times = value->value;
		if (value->value < 0)
		  {
		    error(ERR_NONFATAL, "TIMES value %d is negative",
			  value->value);
		    result->times = 0;
		  }
	      }
	  }
	else
	  {
	    if (result->nprefix == MAXPREFIX)
	      error (ERR_NONFATAL,
		     "instruction has more than %d prefixes", MAXPREFIX);
	    else
	      result->prefixes[result->nprefix++] = tokval.t_integer;
	    i = stdscan(NULL, &tokval);
	  }
      }

    if (i != TOKEN_INSN)
      {
	error (ERR_NONFATAL, "parser: instruction expected");
#ifdef DEBUG
	printf ("parser: instruction expected - got an %d '%c'\n", i, i);
#endif
	result->opcode = -1;
	return result;
      }

    result->opcode = tokval.t_integer;
    
    result->condition = tokval.t_inttwo;

    /*
     * RESB, RESW and RESD cannot be satisfied with incorrectly
     * evaluated operands, since the correct values _must_ be known
     * on the first pass. Hence, even in pass one, we set the
     * `critical' flag on calling evaluate(), so that it will bomb
     * out on undefined symbols. Nasty, but there's nothing we can
     * do about it.
     *
     * For the moment, EQU has the same difficulty, so we'll
     * include that.
     */
    if (result->opcode == I_RESB ||
	result->opcode == I_RESW ||
	result->opcode == I_RESD ||
	result->opcode == I_RESQ ||
	result->opcode == I_REST ||
	result->opcode == I_EQU)
      {
	critical = pass;
      }
    else
      critical = (pass==2 ? 2 : 0);

    if (result->opcode == I_DCB ||
	result->opcode == I_DCW ||
	result->opcode == I_DCD ||
	result->opcode == I_DQ ||
	result->opcode == I_DT ||
	result->opcode == I_INCBIN) 
      {
	extop *eop, **tail = &result->eops, **fixptr;
	int oper_num = 0;

	result->eops_float = FALSE;

	/*
	 * Begin to read the DB/DW/DD/DQ/DT/INCBIN operands.
	 */
	while (1)
	  {
	    i = stdscan(NULL, &tokval);
	    if (i == 0)
	      break;
	    fixptr = tail;
	    eop = *tail = nasm_malloc(sizeof(extop));
	    tail = &eop->next;
	    eop->next = NULL;
	    eop->type = EOT_NOTHING;
	    oper_num++;

	    if (i == TOKEN_NUM && tokval.t_charptr && is_comma_next())
	      {
		eop->type = EOT_DB_STRING;
		eop->stringval = tokval.t_charptr;
		eop->stringlen = tokval.t_inttwo;
		i = stdscan(NULL, &tokval);       /* eat the comma */
		continue;
	      }

	    if ((i == TOKEN_FLOAT && is_comma_next()) || i == '-')
	      {
		long sign = +1L;

		if (i == '-')
		  {
		    char *save = stdscan_bufptr;
		    i = stdscan(NULL, &tokval);
		    sign = -1L;
		    if (i != TOKEN_FLOAT || !is_comma_next())
		      {
			stdscan_bufptr = save;
			i = tokval.t_type = '-';
		      }
		  }

		if (i == TOKEN_FLOAT)
		  {
		    eop->type = EOT_DB_STRING;
		    result->eops_float = TRUE;
		    if (result->opcode == I_DCD)
		      eop->stringlen = 4;
		    else if (result->opcode == I_DQ)
		      eop->stringlen = 8;
		    else if (result->opcode == I_DT)
		      eop->stringlen = 10;
		    else
		      {
			error(ERR_NONFATAL, "floating-point constant"
			      " encountered in `D%c' instruction",
			      result->opcode == I_DCW ? 'W' : 'B');
			/*
			 * fix suggested by Pedro Gimeno... original line
			 * was:
			 * eop->type = EOT_NOTHING;
			 */
			eop->stringlen = 0;
		      }
		    eop = nasm_realloc(eop, sizeof(extop)+eop->stringlen);
		    tail = &eop->next;
		    *fixptr = eop;
		    eop->stringval = (char *)eop + sizeof(extop);
		    if (eop->stringlen < 4 ||
			!float_const (tokval.t_charptr, sign,
				      (unsigned char *)eop->stringval,
				      eop->stringlen, error))
		      eop->type = EOT_NOTHING;
		    i = stdscan(NULL, &tokval);       /* eat the comma */
		    continue;
		  }
	      }

	    /* anything else */ 
	    {
		expr *value;
		value = evaluate (stdscan, NULL, &tokval, NULL,
				  critical, error, NULL);
		i = tokval.t_type;
		if (!value)
		  {			/* error in evaluator */
		    result->opcode = -1;/* unrecoverable parse error: */
		    return result;	/* ignore this instruction */
		  }
		if (is_unknown(value))
		  {
		    eop->type = EOT_DB_NUMBER;
		    eop->offset = 0;	/* doesn't matter what we put */
		    eop->segment = eop->wrt = NO_SEG;   /* likewise */
		  }
		else if (is_reloc(value))
		  {
		    eop->type = EOT_DB_NUMBER;
		    eop->offset = reloc_value(value);
		    eop->segment = reloc_seg(value);
		    eop->wrt = reloc_wrt(value);
		  }
		else
		  {
		    error (ERR_NONFATAL,
			   "operand %d: expression is not simple"
			   " or relocatable", oper_num);
		  }
	    }

	    /*
	     * We're about to call stdscan(), which will eat the
	     * comma that we're currently sitting on between
	     * arguments. However, we'd better check first that it
	     * _is_ a comma.
	     */
	    if (i == 0)		       /* also could be EOL */
	      break;
	    if (i != ',')
	      {
		error (ERR_NONFATAL, "comma expected after operand %d",
		       oper_num);
		result->opcode = -1;/* unrecoverable parse error: */
		return result;     /* ignore this instruction */
	      }
	  }

	if (result->opcode == I_INCBIN)
	  {
	    /*
	     * Correct syntax for INCBIN is that there should be
	     * one string operand, followed by one or two numeric
	     * operands.
	     */
	    if (!result->eops || result->eops->type != EOT_DB_STRING)
		error (ERR_NONFATAL, "`incbin' expects a file name");
	    else if (result->eops->next &&
		     result->eops->next->type != EOT_DB_NUMBER)
		error (ERR_NONFATAL, "`incbin': second parameter is",
		       " non-numeric");
	    else if (result->eops->next && result->eops->next->next &&
		     result->eops->next->next->type != EOT_DB_NUMBER)
		error (ERR_NONFATAL, "`incbin': third parameter is",
		       " non-numeric");
	    else if (result->eops->next && result->eops->next->next &&
		     result->eops->next->next->next)
		error (ERR_NONFATAL, "`incbin': more than three parameters");
	    else
		return result;
	    /*
	     * If we reach here, one of the above errors happened.
	     * Throw the instruction away.
	     */
	    result->opcode = -1;
	    return result;
	  }
	else /* DB ... */
	  if (oper_num == 0)
	    error (ERR_WARNING|ERR_PASS1,
		   "no operand for data declaration");
	  else
	    result->operands = oper_num;

	return result;
      }

    /* right. Now we begin to parse the operands. There may be up to four
     * of these, separated by commas, and terminated by a zero token. */

    bracket = 0;
    minus = 0;
    
    for (operand = 0; operand < 4; operand++)
      {
	expr *value;		       /* used most of the time */
	int mref;		       /* is this going to be a memory ref? */
	int setsize = 0;

	result->oprs[operand].addr_size = 0;/* have to zero this whatever */
	result->oprs[operand].eaflags = 0;   /* and this */
	result->oprs[operand].opflags = 0;

	i = stdscan(NULL, &tokval);
	if (i == 0)
	  break;			  /* end of operands: get out of here */
	result->oprs[operand].type = 0;   /* so far, no override */

	while (i == TOKEN_SPECIAL)
	  {				  /* size specifiers */
	    switch ((int)tokval.t_integer)
	      {
		case S_ASR:
		case S_LSL:
		case S_LSR:
		case S_ROR:
		case S_RRX:		/* RRX is ROR #0 and not impl like this <-- */
		  if (operand != 3 && operand != 2)
		    error (ERR_NONFATAL, "cannot do that here (award winning error message #3)");
		  else
		    result->oprs[operand].shiftflag = tokval.t_integer;
		  
		  break;
		  
		default:
		  error (ERR_NONFATAL, "invalid operand size specification");
	      }
	    i = stdscan(NULL, &tokval);
	  }

	if (i == '[')
	  {
	    mref = TRUE;
	    bracket = 1;
	    i = stdscan(NULL, &tokval);	    
	  }
	result->oprs[operand].bracket = bracket;
	if (i == '+')
	  {
	    i = stdscan (NULL,&tokval);
	  }
	else if (i == '-')
	  {
	    minus = 1;
	    i = stdscan (NULL,&tokval);
	  }
	if (i == TOKEN_REG)
	  {
	    int r;

	    r = tokval.t_integer;

	    result->oprs[operand].type = 0;
	    result->oprs[operand].type |= REGISTER;
	    switch (r)
	      {
		 case R_CPSR_FLG:
		 case R_SPSR_FLG:
		   result->oprs[operand].type |= FLAG;
		   break;
		 default:
		   result->oprs[operand].type |= BITS32;
	      }
	    result->oprs[operand].basereg = r;
	    result->oprs[operand].minus = minus;

	    i = stdscan(NULL, &tokval);
	  }
	else if (i == '{')
	  {
	    int r;
	    int p = 0;
	    int shut_up = 0;
	    
	    result->oprs[operand].type |= REGLIST;
	    result->oprs[operand].basereg = 0;
	    i = stdscan(NULL, &tokval);

	    while (i != '}')
	      {
		if (i != TOKEN_REG)
		  {
		    error (ERR_NONFATAL, "register expected");
		    r = 0;
		  }
		else
		  {
		    r = regtoval (tokval.t_integer);
		    if ((result->oprs[operand].basereg & r) &&
			 shut_up == 0)
		      {
			char buff[256];
			sprintf (buff, "register specified twice in reglist");
			error (ERR_WARNING|ERR_PASS1,
			       buff);
			shut_up = 1;
		      }
		    result->oprs[operand].basereg |= r;
		    if (p)
		      {
			int low;
			int high;

			if (p > r)
			  {
			    high = p >> 1;	/* don't take p twice */
			    low = r << 1;	/* don't take r twice */
			  }
			else		/* also handles p == r */
			  {
			    high = r >> 1;
			    low = p << 1;
			  }
			while (low <= high)
			  {
			    if ((result->oprs[operand].basereg & low) &&
				 shut_up == 0)
			      {
				char buff[256];
				sprintf (buff, "register specified twice in reglist");
				error (ERR_WARNING|ERR_PASS1,
				       buff);
				shut_up = 1;
			      }
			    result->oprs[operand].basereg |= low;
			    low = low << 1;
			  }
			p = 0;
		      }
		  }
		i = stdscan(NULL, &tokval);

		if (i == '-')
		  {
		    p = r;
		    i = stdscan(NULL, &tokval);
		  }
		if (i == ',')
		  {
		    i = stdscan(NULL, &tokval);
		  }

	      }
	    i = stdscan(NULL, &tokval);

	  }
	else
	  {
	    if (i == '#')			/* immediate operand */
	      {
		i = stdscan(NULL, &tokval);
	      }

	    value = evaluate (stdscan, NULL, &tokval,
			  &result->oprs[operand].opflags,
			  critical, error, &hints);
	    i = tokval.t_type;
	    if (result->oprs[operand].opflags & OPFLAG_FORWARD)
	      {
		result->forw_ref = TRUE;
	      }
	    if (!value)
	      {				/* error in evaluator */
		result->opcode = -1;        /* unrecoverable parse error: */
		return result;	        /* ignore this instruction */
	      }
	
	    if (is_just_unknown(value))
	      {     /* it's immediate but unknown */
		result->oprs[operand].type |= IMMEDIATE;
		result->oprs[operand].type |= BITS32;
		result->oprs[operand].offset = 0;	/* don't care */
		result->oprs[operand].segment = NO_SEG; /* don't care again */
		result->oprs[operand].wrt = NO_SEG;	/* still don't care */
	      } 
	    else if (is_reloc(value))			/* it's immediate */
	      {
		result->oprs[operand].type |= IMMEDIATE;
		result->oprs[operand].type |= BITS32;
		result->oprs[operand].offset = reloc_value(value);
		result->oprs[operand].segment = reloc_seg(value);
		result->oprs[operand].wrt = reloc_wrt(value);
		if (is_simple(value) && reloc_value(value)==1)
		    result->oprs[operand].type |= UNITY;
	      }
	  }

	if (i == ']')
	  {
	    mref = TRUE;
	    bracket = 0;
	    i = stdscan(NULL, &tokval);	    
	  }

	if (i == '!')
	  {
	    result->condition |= C_EXSETFLAG;
	    i = stdscan(NULL, &tokval);
	  }
	if (i == '^')
	  {
	    result->condition |= C_FSETFLAG;
	    i = stdscan(NULL, &tokval);
	  }
	if (i != 0 && i != ',')
	  {
	    error (ERR_NONFATAL, "comma or end of line expected");
	    do
	      {		       /* error recovery */
		i = stdscan(NULL, &tokval);
	      } while (i != 0 && i != ',');
	  }
      }

    if (bracket)
      {
	error (ERR_NONFATAL, "']' expected");
      }
    
    result->operands = operand;       /* set operand count */

    while (operand<4)		       /* clear remaining operands */
	result->oprs[operand++].type = 0;

    /*
     * Transform RESW, RESD, RESQ, REST into RESB.
     */
    switch (result->opcode) {
      case I_RESW: result->opcode=I_RESB; result->oprs[0].offset*=2; break;
      case I_RESD: result->opcode=I_RESB; result->oprs[0].offset*=4; break;
      case I_RESQ: result->opcode=I_RESB; result->oprs[0].offset*=8; break;
      case I_REST: result->opcode=I_RESB; result->oprs[0].offset*=10; break;
    }

    return result;
}

static int is_comma_next (void) 
{
    char *p;
    int i;
    struct tokenval tv;

    p = stdscan_bufptr;
    i = stdscan (NULL, &tv);
    stdscan_bufptr = p;
    return (i == ',' || i == ';' || !i);
}

void cleanup_insn (insn *i) 
{
    extop *e;

    while (i->eops) {
	e = i->eops;
	i->eops = i->eops->next;
	nasm_free (e);
    }
}
